/**
 * Copyright (c) 2013-2017 Lorenzo Bettini.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Lorenzo Bettini - Initial contribution and API
 */
package org.eclipse.xsemantics.dsl.ui.outline;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import java.util.LinkedHashSet;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xsemantics.dsl.xsemantics.AbstractFieldDefinition;
import org.eclipse.xsemantics.dsl.xsemantics.AuxiliaryDescription;
import org.eclipse.xsemantics.dsl.xsemantics.AuxiliaryFunction;
import org.eclipse.xsemantics.dsl.xsemantics.CheckRule;
import org.eclipse.xsemantics.dsl.xsemantics.Description;
import org.eclipse.xsemantics.dsl.xsemantics.JudgmentDescription;
import org.eclipse.xsemantics.dsl.xsemantics.Named;
import org.eclipse.xsemantics.dsl.xsemantics.Rule;
import org.eclipse.xsemantics.dsl.xsemantics.XsemanticsFile;
import org.eclipse.xsemantics.dsl.xsemantics.XsemanticsSystem;
import org.eclipse.xtext.EcoreUtil2;
import org.eclipse.xtext.nodemodel.ICompositeNode;
import org.eclipse.xtext.nodemodel.util.NodeModelUtils;
import org.eclipse.xtext.ui.editor.outline.IOutlineNode;
import org.eclipse.xtext.ui.editor.outline.impl.DefaultOutlineTreeProvider;
import org.eclipse.xtext.ui.editor.outline.impl.DocumentRootNode;
import org.eclipse.xtext.ui.editor.outline.impl.EObjectNode;
import org.eclipse.xtext.util.ITextRegion;
import org.eclipse.xtext.util.concurrent.IUnitOfWork;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;

/**
 * customization of the default outline structure
 */
@SuppressWarnings("all")
public class XsemanticsOutlineTreeProvider extends DefaultOutlineTreeProvider {
  protected void _createChildren(final DocumentRootNode parentNode, final XsemanticsFile it) {
    this.createChildren(parentNode, it.getXsemanticsSystem());
  }

  protected void _createChildren(final IOutlineNode parentNode, final XsemanticsSystem it) {
    this.<AbstractFieldDefinition>createNodesFor(parentNode, it.getFields());
    this.createGroupedNodesForAuxliary(parentNode, it);
    this.createGroupedNodesForJudgements(parentNode, it);
    this.<CheckRule>createNodesFor(parentNode, it.getCheckrules());
    final Function1<AuxiliaryFunction, Boolean> _function = (AuxiliaryFunction it_1) -> {
      AuxiliaryDescription __auxiliaryDescription = it_1.get_auxiliaryDescription();
      return Boolean.valueOf((__auxiliaryDescription == null));
    };
    this.<AuxiliaryFunction>createNodesFor(parentNode, IterableExtensions.<AuxiliaryFunction>filter(it.getAuxiliaryFunctions(), _function));
    final Function1<Rule, Boolean> _function_1 = (Rule it_1) -> {
      JudgmentDescription __judgment = it_1.get_judgment();
      return Boolean.valueOf((__judgment == null));
    };
    this.<Rule>createNodesFor(parentNode, IterableExtensions.<Rule>filter(it.getRules(), _function_1));
  }

  private <T extends EObject> void createNodesFor(final IOutlineNode parentNode, final Iterable<T> items) {
    for (final T item : items) {
      this.createNode(parentNode, item);
    }
  }

  private void createGroupedNodesForAuxliary(final IOutlineNode parentNode, final XsemanticsSystem it) {
    final LinkedHashSet<AuxiliaryDescription> rootNodes = CollectionLiterals.<AuxiliaryDescription>newLinkedHashSet();
    rootNodes.addAll(it.getAuxiliaryDescriptions());
    final Function1<AuxiliaryFunction, AuxiliaryDescription> _function = (AuxiliaryFunction it_1) -> {
      return it_1.get_auxiliaryDescription();
    };
    Iterables.<AuxiliaryDescription>addAll(rootNodes, IterableExtensions.<AuxiliaryDescription>filterNull(ListExtensions.<AuxiliaryFunction, AuxiliaryDescription>map(it.getAuxiliaryFunctions(), _function)));
    this.<AuxiliaryDescription>createNodesFor(parentNode, rootNodes);
  }

  private void createGroupedNodesForJudgements(final IOutlineNode parentNode, final XsemanticsSystem it) {
    final LinkedHashSet<JudgmentDescription> rootNodes = CollectionLiterals.<JudgmentDescription>newLinkedHashSet();
    rootNodes.addAll(it.getJudgmentDescriptions());
    final Function1<Rule, JudgmentDescription> _function = (Rule it_1) -> {
      return it_1.get_judgment();
    };
    Iterables.<JudgmentDescription>addAll(rootNodes, IterableExtensions.<JudgmentDescription>filterNull(ListExtensions.<Rule, JudgmentDescription>map(it.getRules(), _function)));
    this.<JudgmentDescription>createNodesFor(parentNode, rootNodes);
  }

  protected void _createNode(final IOutlineNode parentNode, final AuxiliaryDescription it) {
    final EObjectNode node = this.createEObjectNode(parentNode, it);
    final Function1<XsemanticsSystem, Iterable<AuxiliaryFunction>> _function = (XsemanticsSystem system) -> {
      final Function1<AuxiliaryFunction, Boolean> _function_1 = (AuxiliaryFunction function) -> {
        AuxiliaryDescription __auxiliaryDescription = function.get_auxiliaryDescription();
        return Boolean.valueOf(Objects.equal(__auxiliaryDescription, it));
      };
      return IterableExtensions.<AuxiliaryFunction>filter(system.getAuxiliaryFunctions(), _function_1);
    };
    final Function1<XsemanticsSystem, Iterable<AuxiliaryFunction>> childrenProvider = _function;
    node.setTextRegion(this.calculateGroupedTextRegion(node, it, childrenProvider));
  }

  protected void _createNode(final IOutlineNode parentNode, final JudgmentDescription it) {
    final EObjectNode node = this.createEObjectNode(parentNode, it);
    final Function1<XsemanticsSystem, Iterable<Rule>> _function = (XsemanticsSystem system) -> {
      final Function1<Rule, Boolean> _function_1 = (Rule rule) -> {
        JudgmentDescription __judgment = rule.get_judgment();
        return Boolean.valueOf(Objects.equal(__judgment, it));
      };
      return IterableExtensions.<Rule>filter(system.getRules(), _function_1);
    };
    final Function1<XsemanticsSystem, Iterable<Rule>> childrenProvider = _function;
    node.setTextRegion(this.calculateGroupedTextRegion(node, it, childrenProvider));
  }

  private ITextRegion calculateGroupedTextRegion(final IOutlineNode node, final Description it, final Function1<? super XsemanticsSystem, ? extends Iterable<? extends Named>> childrenProvider) {
    ITextRegion _xblockexpression = null;
    {
      ITextRegion _xifexpression = null;
      boolean _isLocalElement = this.isLocalElement(node, it);
      if (_isLocalElement) {
        _xifexpression = node.getFullTextRegion();
      } else {
        _xifexpression = ITextRegion.EMPTY_REGION;
      }
      final ITextRegion parentRegion = _xifexpression;
      XsemanticsSystem xsemanticsSystem = this.getXsemanticsSystem(node.getParent(), it);
      _xblockexpression = this.mergeRegions(parentRegion, childrenProvider.apply(xsemanticsSystem));
    }
    return _xblockexpression;
  }

  private ITextRegion mergeRegions(final ITextRegion startRegion, final Iterable<? extends EObject> items) {
    ITextRegion _xblockexpression = null;
    {
      ITextRegion region = startRegion;
      for (final EObject eo : items) {
        {
          final ICompositeNode parserNode = NodeModelUtils.getNode(eo);
          if ((parserNode != null)) {
            region = region.merge(parserNode.getTextRegion());
          }
        }
      }
      _xblockexpression = region;
    }
    return _xblockexpression;
  }

  protected void _createChildren(final IOutlineNode parentNode, final AuxiliaryDescription it) {
    final XsemanticsSystem xsemanticsSystem = this.getXsemanticsSystem(parentNode.getParent(), it);
    final Function1<AuxiliaryFunction, Boolean> _function = (AuxiliaryFunction function) -> {
      AuxiliaryDescription __auxiliaryDescription = function.get_auxiliaryDescription();
      return Boolean.valueOf(Objects.equal(__auxiliaryDescription, it));
    };
    this.<AuxiliaryFunction>createNodesFor(parentNode, IterableExtensions.<AuxiliaryFunction>filter(xsemanticsSystem.getAuxiliaryFunctions(), _function));
  }

  protected void _createChildren(final IOutlineNode parentNode, final JudgmentDescription it) {
    final XsemanticsSystem xsemanticsSystem = this.getXsemanticsSystem(parentNode.getParent(), it);
    final Function1<Rule, Boolean> _function = (Rule rule) -> {
      JudgmentDescription __judgment = rule.get_judgment();
      return Boolean.valueOf(Objects.equal(__judgment, it));
    };
    this.<Rule>createNodesFor(parentNode, IterableExtensions.<Rule>filter(xsemanticsSystem.getRules(), _function));
  }

  protected boolean _isLeaf(final Rule rule) {
    return true;
  }

  protected boolean _isLeaf(final CheckRule rule) {
    return true;
  }

  protected boolean _isLeaf(final JudgmentDescription desc) {
    return false;
  }

  protected boolean _isLeaf(final AuxiliaryDescription desc) {
    return false;
  }

  protected boolean _isLeaf(final AuxiliaryFunction aux) {
    return true;
  }

  protected boolean _isLeaf(final AbstractFieldDefinition f) {
    return true;
  }

  private XsemanticsSystem getXsemanticsSystem(final IOutlineNode parentNode, final Description description) {
    XsemanticsSystem _xifexpression = null;
    boolean _isLocalElement = this.isLocalElement(parentNode, description);
    if (_isLocalElement) {
      _xifexpression = EcoreUtil2.<XsemanticsSystem>getContainerOfType(description, XsemanticsSystem.class);
    } else {
      _xifexpression = this.getXsemanticsSystem(parentNode);
    }
    return _xifexpression;
  }

  private XsemanticsSystem getXsemanticsSystem(final IOutlineNode it) {
    final IUnitOfWork<XsemanticsSystem, EObject> _function = (EObject eo) -> {
      return EcoreUtil2.<XsemanticsSystem>getContainerOfType(eo, XsemanticsSystem.class);
    };
    return it.<XsemanticsSystem>readOnly(_function);
  }
}
